# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024-  Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass


@dataclass
class ConfigurationError(Exception):
    name: str

    def __str__(self) -> str:
        return f"'{self.name}'"

    def __format__(self, __format_spec: str) -> str:
        match __format_spec:
            case "class":
                return self.__class__.__name__
            case _:
                return f"{self.__class__.__name__}({str(self)})"


class MissingConfiguration(ConfigurationError):
    pass


@dataclass
class InvalidConfiguration(ConfigurationError):
    value: str

    def __str__(self) -> str:
        return f"{super().__str__()}, '{self.value}'"


class ConfigurationFileUnreadable(InvalidConfiguration):
    pass



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024-  Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum
import os
import pprint
from dataclasses import dataclass, field
from datetime import date
from typing import Any, Self
from unittest.mock import patch
from uuid import UUID, uuid4

from envenom import config, optional, required, subconfig, with_default
from envenom.parsers import as_boolean

# some custom classes with different structure and behavior


@dataclass
class CallMe:
    # CallMe is callable, so we need to wrap it in a closure
    # to ensure the instance will get injected when defining the field

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __call__(self, *_: Any, **__: Any) -> Self:
        raise RuntimeError("this is not the desired result!")


class ExitCode(enum.IntEnum):
    OK = 0
    MISSING_CONFIG = 1
    INVALID_CONFIG = 2
    CONFIG_FILE_UNREADABLE = 3


class LaunchCode(enum.StrEnum):
    OK = enum.auto()
    LAUNCHPAD_OBSTRUCTED = enum.auto()
    NOT_ENOUGH_FUEL = enum.auto()
    OVERRIDDEN_BY_POTUSPL = enum.auto()


@dataclass
class BoxedInt:
    value: int

    @classmethod
    def from_str(cls, s: str) -> Self:
        return cls(int(s))


# and our config section


@config(namespace="ns1")
class BoxedIntsSubconfig:
    some_boxed_int_1: BoxedInt = with_default(BoxedInt.from_str, default=BoxedInt(17))
    some_boxed_int_2: BoxedInt = required(BoxedInt.from_str)


@config(namespace=("ns1", "sub-ns1"))
class NamespacedSubconfig:
    some_int: int | None = optional(int)
    some_int_enum: ExitCode | None = optional(lambda c: ExitCode(int(c)))
    some_str_enum: LaunchCode | None = optional(LaunchCode)


@config()
class NonNamespacedSubconfig:
    some_str: str | None = optional()
    some_uuid_1: UUID = with_default(UUID, default=uuid4)
    some_uuid_2: UUID = with_default(UUID, default=uuid4)


@dataclass
class FrameworkConfig:
    is_month_number_odd: bool = field(
        default_factory=lambda: date.today().month % 2 == 0
    )
    feature_flag: bool = True
    boxed_ints: BoxedIntsSubconfig = subconfig(BoxedIntsSubconfig)


@config(namespace="ns1")
class ApplicationConfig:
    some_str: str = required()
    some_bool: bool = required(as_boolean)
    good_time: CallMe = with_default(lambda _: CallMe(), default=lambda: CallMe())
    namespaced_subconfig: NamespacedSubconfig = subconfig(NamespacedSubconfig)
    non_namespaced_subconfig: NonNamespacedSubconfig = subconfig(NonNamespacedSubconfig)
    framework_config: FrameworkConfig = subconfig(FrameworkConfig)


# run it

if __name__ == "__main__":
    env = {
        # no namespace
        "SOME_UUID_1": "9f4d52c5-6a9c-47ed-9cee-7dbbf0e17499",
        # namespace `ns1`
        "NS1__SOME_BOXED_INT_2": "420",
        "NS1__SOME_STR": "pythagoras",
        "NS1__SOME_BOOL": "TruE",
        # namespace `ns1`, `subns1`
        "NS1__SUB_NS1__SOME_INT": "69",
        "NS1__SUB_NS1__SOME_INT_ENUM": "0",
        "NS1__SUB_NS1__SOME_STR_ENUM": "overridden_by_potuspl",
    }
    with patch.dict(os.environ, env):
        pprint.pprint(ApplicationConfig())



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024-  Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# flake8: noqa

from functools import cached_property

from envenom import config, optional, required, subconfig, with_default
from envenom.parsers import as_boolean


@config(namespace=("myapp", "postgres"))
class DatabaseCfg:
    host: str = required()
    port: int = with_default(int, default=5432)
    database: str = required()
    username: str | None = optional()
    password: str | None = optional()
    connection_timeout: int | None = optional(int)
    sslmode_require: bool = with_default(as_boolean, default=False)

    @cached_property
    def connection_string(self) -> str:
        auth = f"{self.username}:{self.password}" if self.password else self.username

        query: dict[str, str] = {}

        if self.connection_timeout:
            query["timeout"] = str(self.connection_timeout)
        if self.sslmode_require:
            query["sslmode"] = "require"

        query_string = "&".join({f"{key}={value}" for key, value in query.items()})

        return (
            f"postgresql+psycopg://{auth}@{self.host}:{self.port}"
            f"/{self.database}?{query_string}"
        )


@config(namespace="myapp")
class AppCfg:
    secret_key: str = required()
    database: DatabaseCfg = subconfig(DatabaseCfg)


if __name__ == "__main__":
    cfg = AppCfg()

    # fmt: off
    print(f"myapp/secret_key: {repr(cfg.secret_key)} {type(cfg.secret_key)}")
    print(f"myapp/db/host: {repr(cfg.database.host)} {type(cfg.database.host)}")
    print(f"myapp/db/port: {repr(cfg.database.port)} {type(cfg.database.port)}")
    print(f"myapp/db/database: {repr(cfg.database.database)} {type(cfg.database.database)}")
    print(f"myapp/db/username: {repr(cfg.database.username)} {type(cfg.database.username)}")
    print(f"myapp/db/password: {repr(cfg.database.password)} {type(cfg.database.password)}")
    print(f"myapp/db/connection_timeout: {repr(cfg.database.connection_timeout)} {type(cfg.database.connection_timeout)}")
    print(f"myapp/db/sslmode_require: {repr(cfg.database.sslmode_require)} {type(cfg.database.sslmode_require)}")
    print(f"myapp/db/connection_string: {repr(cfg.database.connection_string)} {type(cfg.database.connection_string)}")
    # fmt: on



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024-  Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Generic, TypeVar

from envenom.vars import (
    DefaultType,
    Namespace,
    OptionalVar,
    ParserType,
    RequiredVar,
    Var,
    VarWithDefault,
)

T = TypeVar("T")


@dataclass
class Entry(Generic[T], metaclass=ABCMeta):
    parser: ParserType[T] = field(kw_only=True)
    file: bool = field(kw_only=True, default=True)

    @abstractmethod
    def get_var(self, name: str, namespace: Namespace | None) -> Var[T]: ...


class OptionalEntry(Entry[T]):
    def get_var(self, name: str, namespace: Namespace | None) -> OptionalVar[T]:
        return OptionalVar(name, namespace, parser=self.parser, file=self.file)


@dataclass
class EntryWithDefault(Entry[T]):
    default: DefaultType[T] = field(kw_only=True)

    def get_var(self, name: str, namespace: Namespace | None) -> VarWithDefault[T]:
        return VarWithDefault(
            name,
            namespace,
            parser=self.parser,
            file=self.file,
            default=self.default,
        )


class RequiredEntry(Entry[T]):
    def get_var(self, name: str, namespace: Namespace | None) -> RequiredVar[T]:
        return RequiredVar(name, namespace, parser=self.parser, file=self.file)



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024-  Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections.abc import Callable, Iterable
from dataclasses import Field, dataclass, field
from typing import Any, Type, TypeVar

from envenom.entries import Entry, EntryWithDefault, OptionalEntry, RequiredEntry
from envenom.vars import DefaultType, ParserType, Var

T = TypeVar("T")

# Why so much 'type: ignore' in this file?
#
# It's a hard truth to accept, but dataclasses are lying to us.
#
# When the `field` function is called it creates a Field[T] object, but it tells
# us that it really returns a T (or sometimes None), I swear, pretty please.
#
# Because we're exposing a similar interface and hooking into this exact layer,
# we therefore need to:
#   - (1) ignore the return value from `field`; it's a lie anyway
#   - (2) lie to our consumers on the API side just like `field` would, except of
#         course we supply three versions of the `field` function, so we need
#         to lie on all three fronts.


def __default_parser(v: str) -> str:
    """
    This is technically not strictly needed and could be inlined in Var code.
    But the nice bonus is that doing it this way we coerce the default return type
    to `str` if the user does not specify any parser.
    """
    return v


def config(
    namespace: Iterable[str] | str | None = None,
) -> Callable[[Type[T]], Type[T]]:
    def wrapper(cls: Type[T]) -> Type[T]:
        field_var: Var[Any]
        new_fields: dict[str, Field[Any]] = {
            field_name: field(
                init=False,
                repr=True,
                hash=True,
                metadata={
                    "type": cls.__annotations__[field_name],
                    "var": (field_var := field_entry.get_var(field_name, namespace)),
                },
                default_factory=field_var.get,  # type: ignore (1)
            )
            for field_name, field_entry in cls.__dict__.items()
            if isinstance(field_entry, Entry) and field_name in cls.__annotations__
        }

        for field_name, field_entry in new_fields.items():
            setattr(cls, field_name, field_entry)

        return dataclass(frozen=True, eq=True)(cls)

    return wrapper


def subconfig(factory: Callable[[], T]) -> T:
    return field(default_factory=factory)


def optional(
    parser: ParserType[T] = __default_parser, *, file: bool = True
) -> T | None:
    return OptionalEntry(parser=parser, file=file)  # type: ignore (2)


def with_default(
    parser: ParserType[T] = __default_parser,
    *,
    default: DefaultType[T],
    file: bool = True,
) -> T:
    return EntryWithDefault(
        parser=parser,
        file=file,
        default=default,  # type: ignore (2)
    )


def required(parser: ParserType[T] = __default_parser, *, file: bool = True) -> T:
    return RequiredEntry(parser=parser, file=file)  # type: ignore (2)



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024-  Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


def as_boolean(v: str) -> bool:
    """
    This handles one of the most basic conversions in a sensible manner.
    `bool`, unlike `int`, will not do the right thing, so here's a convenience function.

    `t`, `true`, `y`, `yes`, `1`, `+`, `✓` will evaluate to `True`.

    `f`, `false`, `n`, `no`, `0`, `-`, `✗` will evaluate to `False`.

    The parser is case-insensitive.
    """
    normalized = v.lower()
    if normalized in {"t", "true", "y", "yes", "1", "+", "✓"}:
        return True
    if normalized in {"f", "false", "n", "no", "0", "-", "✗"}:
        return False
    raise ValueError(v)
